/* -*- mode:prolog; -*-
 ******************************************************************
 * $Id$
 * Copyright (C) 2008 by INRIA 
 * Author: Eric de la Clergerie <Eric.De_La_Clergerie@inria.fr>
 * ----------------------------------------------------------------
 *
 *  chart.pl -- une implementation d'un algorithme a charte
 *
 * ----------------------------------------------------------------
 * Description
 * une implementation d'un algorithme a charte de type Earley
 * ----------------------------------------------------------------
 */

%% Bibliothque pour les impressions
:-require 'format.pl'.

%% Bibliotheque pour les mta-call
%% :-require 'call.pl'.

:-extensional 'C'/3.

'C'([T|R],T,R).

:-extensional (-->)/2.
:-extensional verbose_opt.

:-rec_prolog verbose/2.

verbose(Msg,Args) :-
   argv(L),
   ( verbose_opt ->
       format(Msg,Args)
   ;   
       true
   ).

:-rec_prolog earley_parse/4.

earley_parse(Goal,L,R,ParseTree) :-
  predict(Goal(L),[],Agenda),
  process(Agenda),
  stored(Answer::(Goal --> true)(L,R)),
  forest(Answer,ParseTree),
  true
  .

:-rec_prolog process/1.

process([]).
process([Item|OldAgenda]) :-
	verbose('process ~w\n',[Item]),
	process_item(Item,OldAgenda,Agenda),
	process(Agenda).

:-rec_prolog process_item/3.

process_item(Item::(A-->Body)(I,J),OldAgenda,Agenda) :-
    ( Body = true ->
      verbose('  try passive\n',[]),
      resolve_passive(A(I,J),OldAgenda,Agenda)
    ; Body = (B,Beta) ->
      verbose('  try active\n',[]),
      predict(B(J),OldAgenda,MiddleAgenda),
      resolve_active(Item,MiddleAgenda,Agenda)
    ).

:-rec_prolog predict/3.

predict(G::A(I),Agenda1,Agenda2) :-
	( A = [_|_] ->
	    Agenda1=Agenda2
	; A = {_} ->
	    Agenda1=Agenda2
	;
	  all_solutions(NewItem,
			G^prediction(G,NewItem),
			Agenda1,
			Agenda2)
	).

:-rec_prolog prediction/2.

prediction(A(I),NewItem) :-
	recorded((A --> Body)),
	dcg_normalize(Body,XBody),
	store(NewItem::(A --> XBody)(I,I),call).

:-rec_prolog resolve_passive/3.

resolve_passive(Fact::A(J,K),Agenda1,Agenda2) :-
  all_solutions(NewItem,
               Fact^passive(Fact,NewItem),
               Agenda1,
               Agenda2).

:-rec_prolog passive/2.

passive(B(J,K),NewItem) :-
	stored(Trans::(A --> (B,Beta))(I,J)),
	store(NewItem::(A --> Beta)(I,K),and(Trans,(B --> true)(J,K))).

:-std_prolog resolve_active/3.
	 
resolve_active(Item::(A-->B,Beta)(I,J),Agenda1,Agenda2) :-
     all_solutions(NewItem,
		   Item^active(Item,NewItem),
		   Agenda1,
		   Agenda2).

:-rec_prolog active/2.

active(Trans::(A-->B,Beta)(I,J),NewItem) :-
   stored(Item::(B --> true)(J,K)),
   store(NewItem::(A --> Beta)(I,K),and(Trans,Item)).

active(Trans::(A-->B::[_|_],Beta)(I,J),NewItem) :-
   scan(B,J,K),
   store(NewItem::(A --> Beta)(I,K),and(Trans,scan(B))).

active(Trans::(A-->B::{G},Beta)(I,J),NewItem) :-
     format('Trying escape on ~w\n',[G]),
%%     call!pcall(G),
     G,
     format('=> Got ~w\n',[G]),
     store(NewItem::(A --> Beta)(I,J),and(Trans,escape(G))).

:-rec_prolog scan/3.

scan([],L,L).
scan([Token|Tokens],L,R) :-
	'C'(L,Token,M),
	scan(Tokens,M,R).

:-rec_prolog stored/1.

stored(Item) :-
	recorded(item(Item,_)),
	true
	.

:-rec_prolog store/2.

store(Item,BackPtr) :-
	( recorded(I::item(Item,M)) ->
	  mutable_read(M,BackPtr1),
	  mutable(M,or(BackPtr1,BackPtr)),
	  fail
	;
	  verbose('\tregistering ~w\n',[Item]),
	  mutable(M,BackPtr,false),
	  record(I)
	).

:-rec_prolog dcg_normalize/2.

dcg_normalize(Body,XBody) :-
	( Body = (B,Rest) ->
	    dcg_normalize(Rest,XRest),
	    XBody = (B,XRest)
	;
	    XBody = (Body,true)
	).

:-xcompiler((all_solutions(NewItem,Item^Goal,Agenda1,Agenda2)
	    :-
	    mutable(M,Agenda1),
	     every((Goal,
		    mutable_read(M,_A),
		    mutable(M,[NewItem|_A]))),
	     mutable_read(M,Agenda2),
	     true
	    )).


:-rec_prolog get_sentence/1.
get_sentence(Sentence) :-
	argv(L),
	( L=['-verbose'|Sentence] -> record(verbose_opt) ;  L = Sentence).

%% La requte,
%% ./chart french.dcg -- Paul mange une pomme
%% ./chart french.dcg -- -verbose Paul mange une pomme
?- get_sentence(Sentence),
    earley_parse(s,Sentence,[],ParseTree).

:-rec_prolog lexicon/2.

:-public(lexicon/2).

%% lexicon(Form,Struct)


%% Pour les forts
:-light_tabular forest/2.
:-mode(forest/2,+(+,-)).

forest(Item,Tree) :-
%%	format('try forest ~w\n',[Item]),
	( Item = scan(Token) ->
	  Tree = Token
	; Item = escape(G) ->
	  Tree = {G}
	; Item = (A --> true)(I,J),
	  recorded( item(Item,M) ),
	  mutable_read(M,BackPtr),
	  backptr(BackPtr,[],Children),
	  Tree =.. [A|Children]
	),
%%	format('forest ~w => ~w\n',[Item,Tree]),
	true
	.

:-std_prolog backptr/3.
backptr(BackPtr,ChildIn,ChildOut) :-
%%	format('try backptr ~w in=~w\n',[BackPtr,ChildIn]),
	( BackPtr = or(BP1,BP2) ->
	  ( backptr(BP1,ChildIn,ChildOut)
	  ;
	    backptr(BP2,ChildIn,ChildOut)
	  )
	; BackPtr = and(TransParent,ItemParent) ->
	  forest(ItemParent,Tree),
	  recorded( item(TransParent,M) ),
	  mutable_read(M,BP),
	  backptr(BP,[Tree|ChildIn],ChildOut)
	; BackPtr = call,
	  ChildIn=ChildOut
	),
%%	format('backptr ~w in=~w => ChildOut\n',[BackPtr,ChildIn,ChildOut]),
	true
	.



	
